#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""LI.FI Bridge provider."""


import enum
import time
import typing as t
from http import HTTPStatus
from math import ceil
from urllib.parse import urlencode

import requests
from autonomy.chain.base import registry_contracts

from operate.bridge.providers.bridge_provider import (
    BridgeProvider,
    BridgeRequest,
    BridgeRequestStatus,
    DEFAULT_MAX_QUOTE_RETRIES,
    GAS_ESTIMATE_BUFFER,
    MESSAGE_QUOTE_ZERO,
    QuoteData,
)
from operate.constants import ZERO_ADDRESS
from operate.operate_types import Chain


LIFI_DEFAULT_ETA = 5 * 60


class LiFiTransactionStatus(str, enum.Enum):
    """LI.FI transaction status."""

    NOT_FOUND = "NOT_FOUND"
    INVALID = "INVALID"
    PENDING = "PENDING"
    DONE = "DONE"
    FAILED = "FAILED"
    UNKNOWN = "UNKNOWN"

    def __str__(self) -> str:
        """__str__"""
        return self.value


class LiFiBridgeProvider(BridgeProvider):
    """LI.FI Bridge provider."""

    def description(self) -> str:
        """Get a human-readable description of the bridge provider."""
        return "LI.FI Bridge & DEX Aggregation Protocol https://li.fi/"

    def quote(self, bridge_request: BridgeRequest) -> None:
        """Update the request with the quote."""
        self._validate(bridge_request)

        if bridge_request.status not in (
            BridgeRequestStatus.CREATED,
            BridgeRequestStatus.QUOTE_DONE,
            BridgeRequestStatus.QUOTE_FAILED,
        ):
            raise RuntimeError(
                f"Cannot quote bridge request {bridge_request.id} with status {bridge_request.status}."
            )

        if bridge_request.execution_data:
            raise RuntimeError(
                f"Cannot quote bridge request {bridge_request.id}: execution already present."
            )

        from_chain = bridge_request.params["from"]["chain"]
        from_address = bridge_request.params["from"]["address"]
        from_token = bridge_request.params["from"]["token"]
        to_chain = bridge_request.params["to"]["chain"]
        to_address = bridge_request.params["to"]["address"]
        to_token = bridge_request.params["to"]["token"]
        to_amount = bridge_request.params["to"]["amount"]

        if to_amount == 0:
            self.logger.info(f"[LI.FI BRIDGE] {MESSAGE_QUOTE_ZERO}")
            quote_data = QuoteData(
                bridge_eta=0,
                elapsed_time=0,
                message=MESSAGE_QUOTE_ZERO,
                provider_data=None,
                timestamp=int(time.time()),
            )
            bridge_request.quote_data = quote_data
            bridge_request.status = BridgeRequestStatus.QUOTE_DONE
            return

        url = "https://li.quest/v1/quote/toAmount"
        headers = {"accept": "application/json"}
        params = {
            "fromChain": Chain(from_chain).id,
            "fromAddress": from_address,
            "fromToken": from_token,
            "toChain": Chain(to_chain).id,
            "toAddress": to_address,
            "toToken": to_token,
            "toAmount": to_amount,
            "maxPriceImpact": 0.50,  # TODO determine correct value
        }
        for attempt in range(1, DEFAULT_MAX_QUOTE_RETRIES + 1):
            start = time.time()
            try:
                self.logger.info(f"[LI.FI BRIDGE] GET {url}?{urlencode(params)}")
                response = requests.get(
                    url=url, headers=headers, params=params, timeout=30
                )
                response.raise_for_status()
                response_json = response.json()
                quote_data = QuoteData(
                    bridge_eta=LIFI_DEFAULT_ETA,
                    elapsed_time=time.time() - start,
                    message=None,
                    provider_data={
                        "attempts": attempt,
                        "response": response_json,
                        "response_status": response.status_code,
                    },
                    timestamp=int(time.time()),
                )
                bridge_request.quote_data = quote_data
                bridge_request.status = BridgeRequestStatus.QUOTE_DONE
                return
            except requests.Timeout as e:
                self.logger.warning(
                    f"[LI.FI BRIDGE] Timeout request on attempt {attempt}/{DEFAULT_MAX_QUOTE_RETRIES}: {e}."
                )
                quote_data = QuoteData(
                    bridge_eta=None,
                    elapsed_time=time.time() - start,
                    message=str(e),
                    provider_data={
                        "attempts": attempt,
                        "response": None,
                        "response_status": HTTPStatus.GATEWAY_TIMEOUT,
                    },
                    timestamp=int(time.time()),
                )
            except requests.RequestException as e:
                self.logger.warning(
                    f"[LI.FI BRIDGE] Request failed on attempt {attempt}/{DEFAULT_MAX_QUOTE_RETRIES}: {e}."
                )
                response_json = response.json()
                quote_data = QuoteData(
                    bridge_eta=None,
                    elapsed_time=time.time() - start,
                    message=response_json.get("message") or str(e),
                    provider_data={
                        "attempts": attempt,
                        "response": response_json,
                        "response_status": getattr(
                            response, "status_code", HTTPStatus.BAD_GATEWAY
                        ),
                    },
                    timestamp=int(time.time()),
                )
            except Exception as e:  # pylint:disable=broad-except
                self.logger.warning(
                    f"[LI.FI BRIDGE] Request failed on attempt {attempt}/{DEFAULT_MAX_QUOTE_RETRIES}: {e}."
                )
                quote_data = QuoteData(
                    bridge_eta=None,
                    elapsed_time=time.time() - start,
                    message=str(e),
                    provider_data={
                        "attempts": attempt,
                        "response": None,
                        "response_status": HTTPStatus.INTERNAL_SERVER_ERROR,
                    },
                    timestamp=int(time.time()),
                )
            if attempt >= DEFAULT_MAX_QUOTE_RETRIES:
                self.logger.error(
                    f"[LI.FI BRIDGE] Request failed after {DEFAULT_MAX_QUOTE_RETRIES} attempts."
                )
                bridge_request.quote_data = quote_data
                bridge_request.status = BridgeRequestStatus.QUOTE_FAILED
                return

            time.sleep(2)

    def _get_approve_tx(self, bridge_request: BridgeRequest) -> t.Optional[t.Dict]:
        """Get the approve transaction."""
        self.logger.info(
            f"[LI.FI BRIDGE] Get appprove transaction for bridge request {bridge_request.id}."
        )

        if bridge_request.params["to"]["amount"] == 0:
            return None

        quote_data = bridge_request.quote_data
        if not quote_data:
            return None

        if not quote_data.provider_data:
            return None

        quote = quote_data.provider_data.get("response")
        if not quote:
            return None

        if "action" not in quote:
            return None

        from_token = quote["action"]["fromToken"]["address"]
        if from_token == ZERO_ADDRESS:
            return None

        transaction_request = quote.get("transactionRequest")
        if not transaction_request:
            return None

        from_amount = int(quote["action"]["fromAmount"])
        from_ledger_api = self._from_ledger_api(bridge_request)

        approve_tx = registry_contracts.erc20.get_approve_tx(
            ledger_api=from_ledger_api,
            contract_address=from_token,
            spender=transaction_request["to"],
            sender=transaction_request["from"],
            amount=from_amount,
        )
        approve_tx["gas"] = 200_000  # TODO backport to ERC20 contract as default
        BridgeProvider._update_with_gas_pricing(approve_tx, from_ledger_api)
        BridgeProvider._update_with_gas_estimate(approve_tx, from_ledger_api)
        approve_tx["gas"] = ceil(approve_tx["gas"] * GAS_ESTIMATE_BUFFER)
        return approve_tx

    def _get_bridge_tx(self, bridge_request: BridgeRequest) -> t.Optional[t.Dict]:
        """Get the bridge transaction."""
        self.logger.info(
            f"[LI.FI BRIDGE] Get bridge transaction for bridge request {bridge_request.id}."
        )

        if bridge_request.params["to"]["amount"] == 0:
            return None

        quote_data = bridge_request.quote_data
        if not quote_data:
            return None

        if not quote_data.provider_data:
            return None

        quote = quote_data.provider_data.get("response")
        if not quote:
            return None

        if "action" not in quote:
            return None

        transaction_request = quote.get("transactionRequest")
        if not transaction_request:
            return None

        from_ledger_api = self._from_ledger_api(bridge_request)

        bridge_tx = {
            "value": int(transaction_request["value"], 16),
            "to": transaction_request["to"],
            "data": transaction_request["data"],  # TODO remove bytes?
            "from": transaction_request["from"],
            "chainId": transaction_request["chainId"],
            "gasPrice": int(transaction_request["gasPrice"], 16),
            "gas": int(transaction_request["gasLimit"], 16),
            "nonce": from_ledger_api.api.eth.get_transaction_count(
                transaction_request["from"]
            ),
        }
        BridgeProvider._update_with_gas_pricing(bridge_tx, from_ledger_api)
        BridgeProvider._update_with_gas_estimate(bridge_tx, from_ledger_api)
        return bridge_tx

    def _update_execution_status(self, bridge_request: BridgeRequest) -> None:
        """Update the execution status."""

        if bridge_request.status not in (
            BridgeRequestStatus.EXECUTION_PENDING,
            BridgeRequestStatus.EXECUTION_UNKNOWN,
        ):
            return

        execution_data = bridge_request.execution_data
        if not execution_data:
            raise RuntimeError(
                f"Cannot update bridge request {bridge_request.id}: execution data not present."
            )

        if not execution_data.from_tx_hash:
            return

        url = "https://li.quest/v1/status"
        headers = {"accept": "application/json"}
        params = {
            "txHash": execution_data.from_tx_hash,
        }

        try:
            self.logger.info(f"[LI.FI BRIDGE] GET {url}?{urlencode(params)}")
            response = requests.get(url=url, headers=headers, params=params, timeout=30)
            response_json = response.json()
            lifi_status = response_json.get(
                "status", str(LiFiTransactionStatus.UNKNOWN)
            )
            execution_data.message = response_json.get(
                "substatusMessage", response_json.get("message")
            )
            response.raise_for_status()
        except Exception as e:
            self.logger.error(
                f"[LI.FI BRIDGE] Failed to update bridge status for {bridge_request.id}: {e}"
            )

        if lifi_status == LiFiTransactionStatus.DONE:
            self.logger.info(f"[LI.FI BRIDGE] Execution done for {bridge_request.id}.")
            from_ledger_api = self._from_ledger_api(bridge_request)
            from_tx_hash = execution_data.from_tx_hash
            to_ledger_api = self._to_ledger_api(bridge_request)
            to_tx_hash = response_json.get("receiving", {}).get("txHash")
            execution_data.message = None
            execution_data.to_tx_hash = to_tx_hash
            execution_data.elapsed_time = BridgeProvider._tx_timestamp(
                to_tx_hash, to_ledger_api
            ) - BridgeProvider._tx_timestamp(from_tx_hash, from_ledger_api)
            bridge_request.status = BridgeRequestStatus.EXECUTION_DONE
        elif lifi_status == LiFiTransactionStatus.FAILED:
            bridge_request.status = BridgeRequestStatus.EXECUTION_FAILED
        elif lifi_status == LiFiTransactionStatus.PENDING:
            bridge_request.status = BridgeRequestStatus.EXECUTION_PENDING
        else:
            bridge_request.status = BridgeRequestStatus.EXECUTION_UNKNOWN

    def _get_explorer_link(self, bridge_request: BridgeRequest) -> t.Optional[str]:
        """Get the explorer link for a transaction."""
        if not bridge_request.execution_data:
            return None

        tx_hash = bridge_request.execution_data.from_tx_hash
        if not tx_hash:
            return None

        return f"https://scan.li.fi/tx/{tx_hash}"
